/* Copyright (C) Stephen Chung, 1991-1992.  All rights reserved. */

#include "jwp.h"

#include "idm.h"


static char filename[MAXFILENAMELEN];
static char filespec[MAXFILENAMELEN];
static OFSTRUCT of;
static FILE *Fp;
static FILEOPTIONS *filep;
static int FormatChoice;
static BOOL ConvertLineBreaks, AlignKanjis, OutputLineBreaks;
static int LineLength;
static long int FileLength, StatusSteps, CharRead;
static HWND statushwnd, statushwnd1;

static FILEFORMAT OriginalFormat;
static char *OverwriteMessage;

static BOOL HasText, InWord;
static int LineNumber;
static int BlankSpaces, CharPosition;
static int LastIndent, CurrentIndent, FirstIndent;
static int LastChar;
static BOOL LastTextIsKanji;

static POSITION p;


char *FileExtensions[] = {
    ".*",
    ".JWP",
    ".EUC",
    ".SJS",
    ".JIS",
    ".OLD",
    ".NEC",
    NULL
};

char *FileFormatNames[] = {
	"Auto-Detect",
	"Normal",
	"EUC",
	"Shift JIS",
    "New JIS",
	"Old JIS",
    "NEC JIS",
    NULL
};

BOOL FAR PASCAL FileOpenDlgProc (HWND, WORD, WORD, LONG);
BOOL FAR PASCAL FileSaveAsDlgProc (HWND, WORD, WORD, LONG);
extern BOOL FAR PASCAL FileFindProc (HWND, WORD, WORD, LONG);



static void MoveStatusBar (long int percent)
{
    long int len;
    char     buffer[10];
    HWND     hwnd;
    RECT     rect;
    HDC      hdc;

    if (percent < 0L) return;
    if (percent > 100L) percent = 100L;

    sprintf(buffer, "%ld%%", percent);
    SetDlgItemText(statushwnd, 4204, buffer);

    hwnd = GetDlgItem(statushwnd, 4205);
    GetClientRect(hwnd, &rect);
    len = (rect.right - 4) * percent / 100L;

    hdc = GetDC(hwnd);
    SelectObject(hdc, GetStockObject(GRAY_BRUSH));
    Rectangle(hdc, 2, 2, len + 2, rect.bottom - 2);
    ReleaseDC(hwnd, hdc);
}



void NewParagraph (FILEOPTIONS *f, POSITION *pos, unsigned int n)
{
    POSITION p = *pos;
    PARAGRAPH far *pp;
    ONELINE far *lp;


    if (PARAOF(p) == NULL) {
        PARAOF(p) = f->paragraph = f->eof = pp = StructAlloc(PARAGRAPH);
        pp->prev = pp->next = NULL;
    } else {
        PARAOF(p)->next = f->eof = pp = StructAlloc(PARAGRAPH);
        pp->prev = PARAOF(p);
        pp->next = NULL;
        PARAOF(p) = pp;
    }

	pp->leftindent = pp->firstindent = pp->rightindent = 0;
	n *= TEXTBLOCKSIZE;
    pp->text = (UNIT far *) BlockAlloc(n * sizeof(UNIT));
    pp->textsize = n;
    pp->text[0].kanji = '\0';

    LINEOF(p) = lp = pp->lines = pp->lastline = StructAlloc(ONELINE);
    lp->next = lp->prev = NULL;
    lp->position = 0;
    lp->height = BASEFONT->height;
    POSOF(p) = lp->length = 0;

    f->nr_lines++;

    LineNumber = 0;
    *pos = p;
}



static void WriteNormalFile (FILE *fp, FILEOPTIONS *f)
{
    int i, j, k;
	UNIT far *cp;
	PARAGRAPH far *pp;
	FILEHEADER header;
	FILEPARAHEADER pheader;

	memset(&header, 0, sizeof(FILEHEADER));
	memset(&pheader, 0, sizeof(FILEPARAHEADER));

	header.magic = MAGIC;
    strcpy(header.version, JWPFILEVERSION);
	header.landscape = f->pagesetup.landscape ? 1 : 0;
    header.paragraphs = 0;

    header.lrheader = f->lrheader;
    header.nofirstpage = f->nofirstpage;

	for (i = 0; i < 4; i++) header.margins[i] = (float) f->pagesetup.margins[i];

	for (i = 0; i < NRSUMMARIES && f->summary[i] == NULL; i++);
    if (i < NRSUMMARIES) header.summary = TRUE;

    for (j = 0; j < 4; j++) {
        for (i = 0; i < NRHEADERS && f->header[j][i] == NULL; i++);
        if (i < NRHEADERS) {
            header.headers = TRUE;
            break;
        }
    }


    /* How many paragraphs? */

    for (pp = f->paragraph; pp != NULL; pp = pp->next, header.paragraphs++);

    fwrite (&header, sizeof(FILEHEADER), 1, fp);


    /* Summaries */

    if (header.summary) {
        KANJI buffer[MAXLINELEN];

        for (i = 0; i < NRSUMMARIES; i++) {
            if (f->summary[i] == NULL) {
                j = 0;
                fwrite(&j, sizeof(int), 1, fp);
                continue;
            }

			j = kanjilen(f->summary[i]) + 1;
            if (j >= MAXLINELEN) j = MAXLINELEN - 1;
            fwrite(&j, sizeof(int), 1, fp);
			kanjicpy(buffer, f->summary[i]);
            fwrite(buffer, sizeof(KANJI), j, fp);
        }
    }

    /* Headers */

    if (header.headers) {
        KANJI buffer[MAXLINELEN];

        for (k = 0; k < 4; k++) {
            for (i = 0; i < NRHEADERS; i++) {
                if (f->header[k][i] == NULL) {
                    j = 0;
                    fwrite(&j, sizeof(int), 1, fp);
                    continue;
                }

                j = kanjilen(f->header[k][i]) + 1;
                if (j >= MAXLINELEN) j = MAXLINELEN - 1;
                fwrite(&j, sizeof(int), 1, fp);
                kanjicpy(buffer, f->header[k][i]);
                fwrite(buffer, sizeof(KANJI), j, fp);
            }
        }
    }

    /* Output the file */

	for (CharRead = 0L, pp = f->paragraph; pp != NULL; pp = pp->next) {

        pheader.textsize = unitlen(pp->text) + 1;
        pheader.firstindent = pp->firstindent;
        pheader.leftindent = pp->leftindent;
        pheader.rightindent = pp->rightindent;
        fwrite(&pheader, sizeof(FILEPARAHEADER), 1, fp);

        for (cp = pp->text; cp->kanji; cp++, CharRead++) {
			if (CharRead % StatusSteps == 0)
				MoveStatusBar(CharRead / StatusSteps);

            if (ISKANJI(cp->kanji)) {
                fputc(HIBYTE(cp->kanji) | 0x0080, fp);
                fputc(LOBYTE(cp->kanji) | 0x0080, fp);
            } else {
                fputc(cp->kanji & 0x007f, fp);
            }
		}

        fputc('\n', fp);
	}

	MoveStatusBar(100L);
}



static BOOL ReadNormalFile (FILE *fp, FILEOPTIONS *f)
{
	unsigned int i, j, n;
	int ch;
	int len;
    FILEHEADER header;
    FILEPARAHEADER pheader;
    POSITION p;

    fread (&header, sizeof(FILEHEADER), 1, fp);

	if (header.magic != MAGIC) {
        ErrorMessage(global.hwnd, "%s is not a normal " PROGNAME " file!", f->filename);
        return (FALSE);
    }
    if (strcmp(header.version, JWPFILEVERSION) < 0) {
        ErrorMessage(global.hwnd, "%s is a file created by an earlier version of "
                                     PROGNAME ".  Unfortunately, it cannot be read by "
                                     "the current version.  We are so very sorry...", f->filename);
        return (FALSE);
    }

	for (i = 0; i < 4; i++) f->pagesetup.margins[i] = (double) header.margins[i];
	f->pagesetup.landscape = header.landscape;
    f->lrheader = header.lrheader;
    f->nofirstpage = header.nofirstpage;
	f->linelen = CalcLineLength(f);


    /* Any summary? */

	for (i = 0; i < NRSUMMARIES; i++) f->summary[i] = NULL;

    if (header.summary) {
        KANJI buffer[MAXLINELEN];

        for (i = 0; i < NRSUMMARIES; i++) {
            fread(&n, sizeof(int), 1, fp);
            if (n > 0) {
                fread((char *) buffer, sizeof(KANJI), n, fp);
				f->summary[i] = (KANJI far *) BlockAlloc((n + 2) * sizeof(KANJI));
				kanjicpy(f->summary[i], buffer);
            }
        }
    }

    /* Any headers? */

    for (j = 0; j < 4; j++) for (i = 0; i < NRHEADERS; i++) f->header[j][i] = NULL;

    if (header.headers) {
        KANJI buffer[MAXLINELEN];

        for (j = 0; j < 4; j++) {
            for (i = 0; i < NRHEADERS; i++) {
                fread(&n, sizeof(int), 1, fp);
                if (n > 0) {
                    fread((char *) buffer, sizeof(KANJI), n, fp);
                    f->header[j][i] = (KANJI far *) BlockAlloc((n + 2) * sizeof(KANJI));
                    kanjicpy(f->header[j][i], buffer);
                }
            }
        }
    }


    /* Read the file */

    PARAOF(p) = NULL;
    LINEOF(p) = NULL;
    POSOF(p) = 0;

    CharRead = 0L;

    for (i = 0; i < header.paragraphs; i++) {
        fread(&pheader, sizeof(FILEPARAHEADER), 1, fp);

        len = pheader.textsize;
		n = (len / TEXTBLOCKSIZE) + 1;

		NewParagraph(f, &p, n);

        PARAOF(p)->firstindent = pheader.firstindent;
        PARAOF(p)->leftindent = pheader.leftindent;
        PARAOF(p)->rightindent = pheader.rightindent;

		for (j = 0; j < len; j++) {
			ch = fgetc(fp);
			if (ch & 0x0080) {      /* Kanji */
				PARAOF(p)->text[j].kanji = ((ch & 0x007f) << 8) | (fgetc(fp) & 0x007f);
				CharRead += 2L;
			} else {
				PARAOF(p)->text[j].kanji = ch;
				CharRead++;
			}
            f->nr_bytes++;
			if (CharRead % StatusSteps == 0)
				MoveStatusBar(CharRead / StatusSteps);
		}
        PARAOF(p)->text[len-1].kanji = 0;
        LINEOF(p)->length = len-1;
	}

    MoveStatusBar(100L);
	return (TRUE);
}



static int OpenCharIn (void)
{
    long int percent;
    int ch;

    ch = fgetc(Fp);

    if (statushwnd == NULL) return (ch);

    CharRead++;

    if (CharRead % StatusSteps == 0 || ch == EOF) {
        percent = CharRead / StatusSteps;
        if (ch == EOF || CharRead >= FileLength) percent = 100L;

        MoveStatusBar(percent);
    }

    return (ch);
}


static void OpenCharOut (int ch)
{
    int i, j;
    KANJI k;
    KANJI buf[10];


    if (ch == '\r') return;         /* Ignore */

    /* Leading spaces */

    if (ch <= ' ' && ch != '\n') {
        if (!HasText) {
            CurrentIndent++;
        } else {
            BlankSpaces++;
        }
        return;
    }

    if (ch == '\n') {
        CurrentIndent = 0;
        LineNumber++;

        if (!ConvertLineBreaks) {
            NewParagraph(filep, &p, 1);
            FirstIndent = LastIndent = 0;
            LineNumber = 0;
        } else if (!HasText) {             /* Blank line */
            NewParagraph(filep, &p, 1);
            FirstIndent = LastIndent = -1;
            LineNumber = 2;
        }

        BlankSpaces = 0;

        HasText = FALSE;
        return;
    }

    /* Check indents */

    if (!HasText) {
        switch (LineNumber) {
            case 0:
                /* No need to check the first line */
                FirstIndent = LastIndent = CurrentIndent;
				PARAOF(p)->firstindent = PARAOF(p)->leftindent = FirstIndent / 2;
                break;

            case 1:
                /* The second line */
                LastIndent = CurrentIndent;
                PARAOF(p)->leftindent = LastIndent / 2;
                break;

            default:
                /* The rest */
                if (LastIndent != CurrentIndent) {
                    NewParagraph(filep, &p, 1);

                    FirstIndent = LastIndent = CurrentIndent;
					PARAOF(p)->firstindent = PARAOF(p)->leftindent = FirstIndent / 2;
                }
                break;
        }
    }


    /* Take care of the blank spaces */

    if (BlankSpaces > 0) {
        j = (BlankSpaces + 1) / 2;

        if (!HasText) {
            k = 0x2121;
        } else if (!LastTextIsKanji && (ch & 0x80)) {
            if (AlignKanjis) {
                j = 1;
                k = '\t';
            } else {
                j = BlankSpaces;
                k = ' ';
            }
        } else if (LastTextIsKanji && !(ch & 0x80)) {
            j = BlankSpaces;
            k = ' ';
        } else if (!LastTextIsKanji && !(ch & 0x80)) {
            j = BlankSpaces;
            k = ' ';
        } else {
            k = 0x2121;
        }

        buf[0] = k;
        buf[1] = 0;
        for (i = 0; i < j; i++, POSOF(p)++) InsertString (filep, p, buf, 0);

        BlankSpaces = 0;
    }


    /* Insert the character */

    i = 0;

    if (LastChar != 0) {
        if (!HasText && LineNumber > 0 && !LastTextIsKanji) buf[i++] = ' ';
        if (ch & 0x80) {
            buf[i++] = (LastChar << 8) | (ch & 0x7f);
            LastChar = 0;
            LastTextIsKanji = TRUE;
        } else {
            buf[i++] = LastChar;
            buf[i++] = ch;
            LastChar = 0;
            LastTextIsKanji = FALSE;
        }
    } else if (ch & 0x80) {
        LastChar = (ch & 0x7f);
        return;
    } else {
        if (!HasText && LineNumber > 0) buf[i++] = ' ';
        buf[i++] = ch;
        LastTextIsKanji = FALSE;
    }

    buf[i] = 0;

    InsertString (filep, p, buf, 0);
    HasText = TRUE;
    BlankSpaces = 0;

    POSOF(p) += i;
}



BOOL DoFileOpen (char *pathname)
{
    int         fd;
	FILEFORMAT  incode;
	HCURSOR     hCursor;
    FARPROC     dlgproc;
    FILEOPTIONS *ToClose = NULL;
    char        buffer[256];


    /* Is this the first file opened? */

    if (fileoptions != NULL && fileoptions->next == NULL && !fileoptions->changed) {
        GetWindowText(fileoptions->parent, buffer, 256);
        if (!strcmp(buffer, "Untitled #1")) ToClose = fileoptions;
    }


    if (pathname == NULL) {
        dlgproc = MakeProcInstance (FileOpenDlgProc, hInstance);

        if (!DialogBox(hInstance, "FileOpen", global.hwnd, dlgproc)) {
            FreeProcInstance(dlgproc);
            return (FALSE);
        }

        FreeProcInstance(dlgproc);
	} else {
		strcpy(filename, pathname);
        FormatChoice = FF_UNKNOWN;
    }

    hCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
    ShowCursor(TRUE);

    incode = (FILEFORMAT) FormatChoice;

    LastIndent = CurrentIndent = FirstIndent = LastChar = 0;
	LineNumber = BlankSpaces = 0;
    HasText = FALSE;
    LastTextIsKanji = TRUE;

	fd = OpenFile(filename, &of, OF_READ);
	if (fd < 0) return (FALSE);

    /* Normal format? */

    if (incode == FF_UNKNOWN) {
        unsigned long int magic;

        lseek(fd, 0L, 0);
		read(fd, &magic, sizeof(unsigned long int));
        lseek(fd, 0L, 0);

		if (magic == MAGIC) incode = FF_NORMAL;
    }


    Fp = fdopen(fd, "rb");
    if (Fp == NULL) {
        close(fd);
        return (FALSE);
    }


    SetupIO(OpenCharIn, OpenCharOut);

    fseek(Fp, 0L, 2);
    FileLength = ftell(Fp);
    StatusSteps = (FileLength / 100L) + 1L;
    CharRead = 0L;

	fseek(Fp, 0L, 0);


    /* Open the status dialog box */

    statushwnd1 = CreateDialog(hInstance, "FileStatus", 0, NULL);
    SetDlgItemText(statushwnd1, 4201, "Reading File:");
	sprintf(buffer, "%s  (%0.2fK)", filename, (double) FileLength / 1024.0);
    SetDlgItemText(statushwnd1, 4202, buffer);
    SetDlgItemText(statushwnd1, 4203, "Determining File Format...");
    SetDlgItemText(statushwnd1, 4204, "0%");

    statushwnd = NULL;
    if (incode == FF_UNKNOWN) {
        incode = DetectCodeType();
        if (incode == FF_UNKNOWN) {
            if (Confirmation(statushwnd1, "The file %s is not in any of the "
                                "supported Japanese text formats.  Open it anyway?", filename) != IDYES) {
                fclose(Fp);
                DestroyWindow(statushwnd1);
                return (FALSE);
            }
        }
        fseek(Fp, 0L, 0);
    }
    sprintf(buffer, "File Format: %s", (incode == FF_UNKNOWN) ? "Unknown" : FileFormatNames[incode]);
    statushwnd = statushwnd1;
    SetDlgItemText(statushwnd, 4203, buffer);


    filep = NewFile(FN_NORMAL, incode != FF_NORMAL);
    if (filep == NULL) {
        fclose(Fp);
        ErrorMessage(global.hwnd, "Cannot open any more file window!");
        DestroyWindow(statushwnd);
        return (FALSE);
    }

    if (incode == FF_NORMAL) {
		strcpy(filep->filename, of.szPathName);

		if (!ReadNormalFile(Fp, filep)) {
            FILEOPTIONS *f1;

            fclose(Fp);
            DestroyWindow(statushwnd);

            ShowCursor(FALSE);
            SetCursor(hCursor);

            CloseFile(filep);

            /* Unlink it */

            if (fileoptions == filep) {
                fileoptions = filep->next;
            } else {
                for (f1 = fileoptions; f1 != NULL && f1->next != filep; f1 = f1->next);
                if (f1 == NULL) {
                    ErrorMessage(global.hwnd, "Funny: cannot find FILEOPTIONS record to deallocate");
                } else {
                    f1->next = filep->next;
                }
            }
            return (FALSE);
        }
    } else {
        filep->filename[0] = '\0';

        PARAOF(p) = filep->paragraph;
        LINEOF(p) = filep->paragraph->lines;
        POSOF(p) = 0;

        FileImport(incode);
	}

    fclose(Fp);



    /* Reformat */

    SetDlgItemText(statushwnd1, 4201, "Reformatting...");

    TOPPARA(filep) = CURPARA(filep) = filep->paragraph;
    TOPLINE(filep) = CURLINE(filep) = filep->paragraph->lines;
    CURCHAR(filep) = 0;

	ReformatFile(filep);

    DestroyWindow(statushwnd);


    if (incode != FF_UNKNOWN && incode != FF_NORMAL) {
        filep->fileformat = incode;
        sprintf(buffer, "%s (%s)", filename, FileFormatNames[incode]);
    } else {
        strcpy(buffer, filename);
    }
    strcpy(filep->filename, filename);
    ShowFileWindow(buffer, filep);


    if (ToClose != NULL) {
        ToClose->changed = FALSE;
        SendMessage(global.clienthwnd, WM_MDIDESTROY, ToClose->parent, 0L);
    }

    ShowCursor(FALSE);
    SetCursor(hCursor);

	filep->changed = FALSE;

    PutQuickFile(of.szPathName);
    UpdateQuickFiles();

    return (TRUE);
}



static int SaveCharIn (void)
{
    int i;
	KANJI ch, temp;
	long int percent;

	if (LastChar != 0) {
        ch = LastChar;
		LastChar = 0;
        CharPosition++;
		return (ch);
    } else if (PARAOF(p) == NULL || LINEOF(p) == NULL) {
        ch = EOF;
        goto StatusBar;     /* Put the bar to 100% */
    } if (OutputLineBreaks && !HasText && CharPosition <
            (LineNumber ? PARAOF(p)->leftindent : PARAOF(p)->firstindent) * 2) {
        CharPosition++;
		return (' ');
	} else if (OutputLineBreaks && CharPosition >= LineLength - PARAOF(p)->rightindent) {
		HasText = InWord = FALSE;
        CharPosition = 0;
        LineNumber++;
        return ('\n');
    }


    for (;;) {
        ch = (CHAROF(p, POSOF(p)) & 0x7f7f);

        if (!ch) {
			POSOF(p) = 0;
            PARAOF(p) = PARAOF(p)->next;
			if (PARAOF(p) != NULL) LINEOF(p) = PARAOF(p)->lines;
            else LINEOF(p) = NULL;

            HasText = InWord = FALSE;
            CharPosition = LineNumber = 0;
            return ('\n');
		}

        if (!HasText && (ch <= ' ' || ch == 0x2121)) {
			POSOF(p)++;
            CharRead++;
			continue;
		}

		if (ISKANJI(ch)) {
			if (ch == 0x2121) {
				ch = LastChar = ' ';
			} else {
                LastChar = (LOBYTE(ch) | 0x80);
				ch = (HIBYTE(ch) | 0x80);
			}
            InWord = FALSE;
        } else if (ch <= ' ') {
            ch = ' ';
			InWord = FALSE;
        } else {
			ch &= 0x7f7f;

            if (strchr("-", ch) != NULL) {
                InWord = FALSE;
            } else {
                if (!InWord) {          /* Search for word break */
                    for (i = 1; ; i++) {
                        temp = CHAROF(p, POSOF(p)+i);
                        if (!temp) break;
                        if (IsSmallKana(temp)) break;
                        if (temp <= ' ' || strchr("-", temp) != NULL) break;

                        if (OutputLineBreaks && CharPosition + i >= LineLength - PARAOF(p)->rightindent) {
							HasText = InWord = FALSE;
							CharPosition = 0;
                            LineNumber++;
                            return ('\n');
                        }
                    }
                }
                InWord = TRUE;
            }
        }

		break;
    }

	POSOF(p)++;
    CharPosition++;
    HasText = TRUE;

    if (statushwnd == NULL) return (ch);

    CharRead++;

StatusBar:

	if (CharRead % StatusSteps == 0 || ch == EOF) {
        percent = CharRead / StatusSteps;
		if (ch == EOF || CharRead >= FileLength) percent = 100L;

        MoveStatusBar(percent);
    }
	return (ch);
}



static void SaveCharOut (int ch)
{
    fputc(ch, Fp);
}



BOOL FAR PASCAL FileSaveFormatProc (HWND hwnd, WORD message, WORD wParam, LONG lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            char buffer[100];

            sprintf(buffer, OverwriteMessage, filename, FileFormatNames[OriginalFormat]);
            SetDlgItemText(hwnd, 4211, buffer);
            SetDlgItemText(hwnd, 4201, FileFormatNames[OriginalFormat]);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

        case WM_COMMAND:
            switch (wParam) {
                case IDOK:      OriginalFormat = FF_NORMAL;
                                EndDialog(hwnd, TRUE);
                                return (TRUE);

                case IDCANCEL:  OriginalFormat = FF_UNKNOWN;
                                EndDialog(hwnd, TRUE);
                                return (TRUE);

                case 4201:      EndDialog(hwnd, TRUE);
                                return (TRUE);

                default:        break;
            }
            break;
    }

    return (FALSE);
}



BOOL DoFileSaveAs (BOOL Dialog)
{
	int         fd;
	int         i;
	FILEFORMAT  outcode;
	HCURSOR     hCursor;
    FARPROC     dlgproc;
    char        buffer[256];

    if (global.active == NULL) return (FALSE);

    filep = global.active;

    dlgproc = MakeProcInstance (FileSaveAsDlgProc, hInstance);

    if (Dialog || !filep->filename[0]) {
        if (!DialogBox(hInstance, "FileSaveAs", global.hwnd, dlgproc)) {
            FreeProcInstance(dlgproc);
			return (FALSE);
        }
        FreeProcInstance(dlgproc);
        outcode = (FILEFORMAT) FormatChoice + 1;

        SendMessage(global.hwnd, WM_COMMAND, IDM_FILESUMMARY, 0L);
    } else if (filep->fileformat != FF_NORMAL) {
        BOOL same;

        strcpy(filename, filep->filename);

        /* Change the extension to .JWP */

        for (i = strlen(filename) - 1; i >= 0; i--) {
            if (filename[i] == '\\' || filename[i] == '/' ||
				filename[i] == '.' || filename[i] == ':') break;
        }

        if (i < 0 || filename[i] != '.') {
            strcat(filename, FileExtensions[1]);    /* Normal extension */
            same = FALSE;
        } else if (!strncpy(filename + i, FileExtensions[1], 4)) {
            same = TRUE;
        } else {
            strcpy(filename + i, FileExtensions[1]);
            same = FALSE;
        }


        /* Overwrite with normal format? */

        dlgproc = MakeProcInstance (FileSaveFormatProc, hInstance);
        OriginalFormat = filep->fileformat;

        if (!same) OverwriteMessage = "Save file as %s (Normal Format)?\nOr keep it %s?";
        else OverwriteMessage = "Overwrite %s with Normal Format?\nOr keep it %s?";

        MessageBeep(0);

        DialogBox(hInstance, "FileSaveFormat", global.hwnd, dlgproc);
        FreeProcInstance(dlgproc);

        if (OriginalFormat == FF_UNKNOWN) return (FALSE);

        outcode = OriginalFormat;
        FormatChoice = outcode - 1;
    } else {
        strcpy(filename, filep->filename);
        outcode = filep->fileformat;
        FormatChoice = outcode - 1;
    }

    hCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
    ShowCursor(TRUE);

	LineNumber = LastChar = CharPosition = 0;
	HasText = InWord = FALSE;

	fd = OpenFile(filename, &of, OF_WRITE | OF_CREATE);
	if (fd < 0) return (FALSE);

    Fp = fdopen(fd, "wb");
	if (Fp == NULL) {
        close(fd);
        return (FALSE);
    }

	SetupIO(SaveCharIn, SaveCharOut);

    FileLength = 0;
    StatusSteps = (FileLength / 100L) + 1L;
    CharRead = 0L;

	fseek(Fp, 0L, 0);

    /* Open the status dialog box */

    statushwnd = CreateDialog(hInstance, "FileStatus", 0, NULL);
    SetDlgItemText(statushwnd, 4201, "Writing File:");
	SetDlgItemText(statushwnd, 4202, filename);
	sprintf(buffer, "File Format: %s", FileFormatNames[FormatChoice+1]);
    SetDlgItemText(statushwnd, 4203, buffer);
    SetDlgItemText(statushwnd, 4204, "0%");


    CharRead = 0L;
    FileLength = filep->nr_bytes;

    StatusSteps = (FileLength / 100L) + 1L;


    if (outcode == FF_NORMAL) {
		WriteNormalFile(Fp, filep);
	} else {
        PARAOF(p) = filep->paragraph;
		LINEOF(p) = filep->paragraph->lines;
        POSOF(p) = 0;

        FileExport(outcode);
    }

    fclose(Fp);

    strcpy(filep->filename, filename /*of.szPathName*/);
    filep->fileformat = outcode;

	if (outcode != FF_NORMAL) {
        sprintf(buffer, "%s (%s)", filep->filename, FileFormatNames[outcode]);
    } else {
        strcpy(buffer, filep->filename);
    }
    SetWindowText(filep->parent, buffer);

    DestroyWindow(statushwnd);

    ShowCursor(FALSE);
    SetCursor(hCursor);

	filep->changed = FALSE;

    PutQuickFile(of.szPathName);
    UpdateQuickFiles();

	return (TRUE);
}



BOOL FAR PASCAL FileOpenDlgProc (HWND hwnd, WORD message, WORD wParam, LONG lParam)
{
	int i, j;


	switch (message) {

	case WM_INITDIALOG:
		strcpy(filename,"*.*");
        strcpy(filespec,"*.*");

        SendDlgItemMessage(hwnd, 4201, EM_LIMITTEXT, MAXFILENAMELEN, 0L);
		SendDlgItemMessage(hwnd, 4202, EM_LIMITTEXT, MAXFILENAMELEN, 0L);

        DlgDirList(hwnd, filespec, 4211, 4202, 0x0000);
        DlgDirList(hwnd, "*.*", 4212, 4202, 0xc010);

        SetDlgItemText(hwnd, 4201, filespec);

		for (i = 0; FileFormatNames[i] != NULL; i++)
            SendDlgItemMessage(hwnd, 4221, CB_INSERTSTRING, -1, (LONG) ((LPSTR) FileFormatNames[i]));
        SendDlgItemMessage(hwnd, 4221, CB_SETCURSEL, 0, 0L);

        CheckDlgButton (hwnd, 4222, FALSE);
        CheckDlgButton (hwnd, 4223, TRUE);

        CenterDialogBox(hwnd);

		return (TRUE);

    case WM_CHAR:
        switch (wParam) {
            case '\x1b': SendMessage(hwnd, WM_COMMAND, IDCANCEL, 0L);
                         return (TRUE);
        }
        break;

	case WM_COMMAND:
		switch (wParam) {
			case 4211:
                switch (HIWORD(lParam)) {
					case LBN_SELCHANGE:
					case LBN_DBLCLK:
                        DlgDirSelect(hwnd, filename, 4211);
                        SetDlgItemText(hwnd, 4201, filename);
						if (HIWORD(lParam) == LBN_DBLCLK)
                            SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                        return (TRUE);
                }
                break;

			case 4212: {
				char directory[MAXFILENAMELEN];

				switch (HIWORD(lParam)) {
					case LBN_DBLCLK:
                        DlgDirSelect(hwnd, directory, 4212);

						strcat(directory, "*.*");

                        DlgDirList(hwnd, directory, 4212, 4202, 0xc010);
                        DlgDirList(hwnd, filespec, 4211, 4202, 0x0000);

                        SetDlgItemText(hwnd, 4201, filespec);
						return (TRUE);
				}
				break;
			}

			case 4201:
				if (HIWORD(lParam) == EN_CHANGE)
                    EnableWindow(GetDlgItem(hwnd, IDOK), (BOOL) SendMessage(LOWORD(lParam), WM_GETTEXTLENGTH, 0, 0L));
				return (TRUE);

			case 4221:
                if (HIWORD(lParam) == LBN_SELCHANGE) {
                    i = (WORD) SendMessage(GetDlgItem(hwnd, 4221), CB_GETCURSEL, 0, 0L);
					strcpy(filespec,"*");
                    strcat(filespec, FileExtensions[i]);
                    SetDlgItemText(hwnd, 4201, filespec);
                    SendMessage(hwnd, WM_COMMAND, IDOK, 0L);

                    return (TRUE);
                }
                break;

            case 4231: {    /* File Find */
                FARPROC finddlgproc;

                finddlgproc = MakeProcInstance(FileFindProc, hInstance);
                i = DialogBox(hInstance, "FileFind", hwnd, finddlgproc);
                FreeProcInstance(finddlgproc);

                if (i) EndDialog(hwnd, TRUE);
                return (TRUE);
            }

			case IDOK:
                GetDlgItemText(hwnd, 4201, filename, 80);

				i = strlen(filename);
				j = filename[i-1];

				if (j == '\\' || j == ':') strcat(filename, "*.*");

				if (strchr(filename, '*') != NULL || strchr(filename, '?') != NULL) {
					if (DlgDirList(hwnd, filename, 4211, 4202, 0x0000)) {
                        DlgDirList(hwnd, "*.*", 4212, 4202, 0xc010);
						strcpy(filespec, filename);
                        SetDlgItemText(hwnd, 4201, filespec);
					} else {
						MessageBeep(0);
					}

					return (TRUE);
				}

				strcat(strcat(filename, "\\"), "*.*");

                if (DlgDirList(hwnd, filename, 4211, 4202, 0x0000)) {
                    DlgDirList(hwnd, "*.*", 4212, 4202, 0xc010);
					strcpy(filespec, filename);
                    SetDlgItemText(hwnd, 4201, filespec);
					return (TRUE);
				}

                filename[i] = '\0';

				if (OpenFile(filename, &of, OF_READ | OF_EXIST) == -1) {
                    strcat(filename, FileExtensions[1]);    /* Normal */
                    if (OpenFile(filename, &of, OF_READ | OF_EXIST) == -1) {
						MessageBeep(0);
						return (TRUE);
					}
				}

				OemToAnsi(filename, filename);
                AnsiUpper(filename);

                FormatChoice = (WORD) SendDlgItemMessage(hwnd, 4221, CB_GETCURSEL, 0, 0L);
                ConvertLineBreaks = IsDlgButtonChecked(hwnd, 4222);
				AlignKanjis = IsDlgButtonChecked(hwnd, 4223);

				EndDialog(hwnd, TRUE);
				return (TRUE);

            case IDCANCEL:
				EndDialog(hwnd, FALSE);
				return (TRUE);
		}
	}

	return (FALSE);
}



BOOL FAR PASCAL FileSaveAsDlgProc (HWND hwnd, WORD message, WORD wParam, LONG lParam)
{
    int fd;
	int i, j;
	char buffer[10];


	switch (message) {

	case WM_INITDIALOG:
        sprintf(filename,"*%s",FileExtensions[FF_NORMAL]);
        strcpy(filespec,filename);

        SendDlgItemMessage(hwnd, 4201, EM_LIMITTEXT, MAXFILENAMELEN, 0L);
        SendDlgItemMessage(hwnd, 4202, EM_LIMITTEXT, MAXFILENAMELEN, 0L);

		DlgDirList(hwnd, filespec, 4211, 4202, 0x0000);
        DlgDirList(hwnd, "*.*", 4212, 4202, 0xc010);

        SetDlgItemText(hwnd, 4201, filespec);

        for (i = 1; FileFormatNames[i] != NULL; i++)    /* Skips Auto-Detect */
			SendDlgItemMessage(hwnd, 4221, CB_INSERTSTRING, -1, (LONG) ((LPSTR) FileFormatNames[i]));

        /* Select the Normal format */
        SendDlgItemMessage(hwnd, 4221, CB_SETCURSEL, 0, 0L);

        CheckDlgButton (hwnd, 4222, 0);
        EnableWindow(GetDlgItem(hwnd, 4222), FALSE);
		SetDlgItemText(hwnd, 4223, "");
        EnableWindow(GetDlgItem(hwnd, 4223), FALSE);

        CenterDialogBox(hwnd);

		return (TRUE);

    case WM_CHAR:
        switch (wParam) {
			case '\x1b': SendMessage(hwnd, WM_COMMAND, IDCANCEL, 0L);
						 return (TRUE);
		}
		break;

	case WM_COMMAND:
		switch (wParam) {
			case 4211:
				switch (HIWORD(lParam)) {
                    case LBN_SELCHANGE:
					case LBN_DBLCLK:
                        DlgDirSelect(hwnd, filename, 4211);
                        SetDlgItemText(hwnd, 4201, filename);
						if (HIWORD(lParam) == LBN_DBLCLK)
                            SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
						return (TRUE);
                }
                break;

			case 4212: {
				char directory[MAXFILENAMELEN];

                switch (HIWORD(lParam)) {
					case LBN_DBLCLK:
                        DlgDirSelect(hwnd, directory, 4212);

						strcat(directory, "*.*");

                        DlgDirList(hwnd, directory, 4212, 4202, 0xc010);
                        DlgDirList(hwnd, filespec, 4211, 4202, 0x0000);

                        SetDlgItemText(hwnd, 4201, filespec);
						return (TRUE);
				}
				break;
			}

			case 4201:
				if (HIWORD(lParam) == EN_CHANGE)
					EnableWindow(GetDlgItem(hwnd, IDOK), (BOOL) SendMessage(LOWORD(lParam), WM_GETTEXTLENGTH, 0, 0L));
				return (TRUE);

			case 4221:
                if (HIWORD(lParam) == LBN_SELCHANGE) {
                    i = (WORD) SendMessage(GetDlgItem(hwnd, 4221), CB_GETCURSEL, 0, 0L);
                    if (i != 0) {       /* Normal? */
                        EnableWindow(GetDlgItem(hwnd, 4222), TRUE);
                        EnableWindow(GetDlgItem(hwnd, 4223), IsDlgButtonChecked(hwnd, 4222));
                    } else {
						EnableWindow(GetDlgItem(hwnd, 4222), FALSE);
                        EnableWindow(GetDlgItem(hwnd, 4223), FALSE);
                    }
                    strcpy(filespec,"*");
                    strcat(filespec, FileExtensions[i + 1]);
                    SetDlgItemText(hwnd, 4201, filespec);
                    SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                    return (TRUE);
                }
                break;

			case 4222:
				if (IsDlgButtonChecked(hwnd, 4222)) {
					sprintf(buffer, "%d", global.active->linelen);
                    EnableWindow(GetDlgItem(hwnd, 4223), TRUE);
                    SetDlgItemText(hwnd, 4223, buffer);
				} else {
                    EnableWindow(GetDlgItem(hwnd, 4223), FALSE);
                    SetDlgItemText(hwnd, 4223, "");
				}
				return (TRUE);

            case IDOK:
				GetDlgItemText(hwnd, 4201, filename, 80);

				i = strlen(filename);
				j = filename[i-1];

				if (j == '\\' || j == ':') strcat(filename, "*.*");

				if (strchr(filename, '*') != NULL || strchr(filename, '?') != NULL) {
                    if (DlgDirList(hwnd, filename, 4211, 4202, 0x0000)) {
                        DlgDirList(hwnd, "*.*", 4212, 4202, 0xc010);
                        strcpy(filespec, filename);
                        SetDlgItemText(hwnd, 4201, filespec);
					} else {
						MessageBeep(0);
					}

					return (TRUE);
				}

				strcat(strcat(filename, "\\"), "*.*");

                if (DlgDirList(hwnd, filename, 4211, 4202, 0x0000)) {
                    DlgDirList(hwnd, "*.*", 4212, 4202, 0xc010);
					strcpy(filespec, filename);
                    SetDlgItemText(hwnd, 4201, filespec);
					return (TRUE);
				}

                filename[i] = '\0';

                FormatChoice = (WORD) SendDlgItemMessage(hwnd, 4221, CB_GETCURSEL, 0, 0L);

                strcat(filename, FileExtensions[FormatChoice+1]);
                AnsiUpper(filename);

                if (OpenFile(filename, &of, OF_READ | OF_EXIST) != -1) {
                    switch (Confirmation (hwnd, "File %s already exists.\nOverwrite it?", filename)) {
                        case IDCANCEL: EndDialog(hwnd, FALSE); return(TRUE);
						case IDNO:     return (TRUE);
						case IDYES:    break;
					}
                } else if ((fd = OpenFile(filename, &of, OF_WRITE | OF_CREATE)) == -1) {
                    filename[i] = '\0';
                    if (OpenFile(filename, &of, OF_READ | OF_EXIST) != -1) {
                        switch (Confirmation (hwnd, "File %s already exists.\nOverwrite it?", filename)) {
                            case IDCANCEL: EndDialog(hwnd, FALSE); return(TRUE);
                            case IDNO:     return (TRUE);
                            case IDYES:    break;
                        }
                    } else if ((fd = OpenFile(filename, &of, OF_WRITE | OF_CREATE)) == -1) {
						MessageBeep(0);
						return (TRUE);
                    }
                }

                close(fd);

				OemToAnsi(filename, filename);

                OutputLineBreaks = IsDlgButtonChecked(hwnd, 4222);
				if (OutputLineBreaks) {
                    GetDlgItemText(hwnd, 4223, buffer, 3);
                    LineLength = atoi(buffer);
					if (LineLength <= 0) {
                        ErrorMessage(global.hwnd, "Bad Line Length: %s", buffer);
						sprintf(buffer, "%d", global.active->linelen);
                        SetDlgItemText(hwnd, 4223, buffer);
						return (TRUE);
					}
                    LineLength *= 2;
                }


				EndDialog(hwnd, TRUE);
				return (TRUE);

            case IDCANCEL:
				EndDialog(hwnd, FALSE);
				return (TRUE);
		}
	}

	return (FALSE);
}
